"""
Aria2 API.

This module defines the API class, which makes use of a JSON-RPC client to provide higher-level methods to
interact easily with a remote aria2c process.
"""
import functools
import shutil
import threading
from base64 import b64encode
from pathlib import Path
from typing import Callable, List, Optional

from loguru import logger

from aria2p.client import Client, ClientException
from aria2p.downloads import Download
from aria2p.options import Options
from aria2p.stats import Stats
from aria2p.types import OperationResult, OptionsType, PathOrStr
from aria2p.utils import read_lines


class API:
    """
    A class providing high-level methods to interact with a remote aria2c process.

    This class is instantiated with a reference to a [`Client`][aria2p.client.Client] instance. It then uses this client
    to call remote procedures, or remote methods. While the client methods reflect exactly what aria2c is providing
    through JSON-RPC, this class's methods allow for easier / faster control of the remote process. It also
    wraps the information the client retrieves in Python object, like [`Download`][aria2p.downloads.Download],
    allowing for even more Pythonic interactions, without worrying about payloads, responses, JSON, etc..
    """

    def __init__(self, client: Optional[Client] = None) -> None:
        """
        Initialize the object.

        Arguments:
            client: An instance of the [aria2p.client.Client][] class.
        """
        self.client = client or Client()
        self.listener: Optional[threading.Thread] = None

    def __repr__(self) -> str:
        return f"API({self.client!r})"

    def add(self, uri: str) -> List[Download]:  # noqa: WPS231 (not that complex)
        """
        Add a download (guess its type).

        If the provided URI is in fact a file-path, and is neither a torrent or a metalink,
        then we read its lines and try to add each line as a download, recursively.

        Arguments:
            uri: The URI or file-path to add.

        Returns:
            The created downloads.
        """
        new_downloads = []
        path = Path(uri)

        # On Windows, path.exists() generates an OSError when path is an URI
        # See https://github.com/pawamoy/aria2p/issues/41
        try:
            path_exists = path.exists()
        except OSError:
            path_exists = False

        if path_exists:
            if path.suffix == ".torrent":
                new_downloads.append(self.add_torrent(path))
            elif path.suffix == ".metalink":
                new_downloads.extend(self.add_metalink(path))
            else:
                for line in read_lines(path):
                    if line:
                        new_downloads.extend(self.add(line))
        elif uri.startswith("magnet:?"):
            new_downloads.append(self.add_magnet(uri))
        else:
            new_downloads.append(self.add_uris([uri]))

        return new_downloads

    def add_magnet(self, magnet_uri: str, options: OptionsType = None, position: int = None) -> Download:
        """
        Add a download with a Magnet URI.

        Arguments:
            magnet_uri: The Magnet URI.
            options: An instance of the [`Options`][aria2p.options.Options] class or a dictionary
                containing aria2c options to create the download with.
            position: The position where to insert the new download in the queue. Start at 0 (top).

        Returns:
            The newly created download object.
        """
        if options is None:
            options = {}

        if isinstance(options, Options):
            client_options = options.get_struct()
        else:
            client_options = options

        gid = self.client.add_uri([magnet_uri], client_options, position)

        return self.get_download(gid)

    def add_torrent(
        self,
        torrent_file_path: PathOrStr,
        uris: List[str] = None,
        options: OptionsType = None,
        position: int = None,
    ) -> Download:
        """
        Add a download with a torrent file (usually .torrent extension).

        Arguments:
            torrent_file_path: The path to the torrent file.
            uris: A list of URIs used for Web-seeding.
            options: An instance of the [`Options`][aria2p.options.Options] class or a dictionary
                containing aria2c options to create the download with.
            position: The position where to insert the new download in the queue. Start at 0 (top).

        Returns:
            The newly created download object.
        """
        if uris is None:
            uris = []

        if options is None:
            options = {}

        if isinstance(options, Options):
            client_options = options.get_struct()
        else:
            client_options = options

        with open(torrent_file_path, "rb") as stream:
            torrent_contents = stream.read()
        encoded_contents = b64encode(torrent_contents).decode("utf8")

        gid = self.client.add_torrent(encoded_contents, uris, client_options, position)

        return self.get_download(gid)

    def add_metalink(
        self,
        metalink_file_path: PathOrStr,
        options: OptionsType = None,
        position: int = None,
    ) -> List[Download]:
        """
        Add a download with a Metalink file.

        Arguments:
            metalink_file_path: The path to the Metalink file.
            options: An instance of the [`Options`][aria2p.options.Options] class or a dictionary
                containing aria2c options to create the download with.
            position: The position where to insert the new download in the queue. Start at 0 (top).

        Returns:
            The newly created download objects.
        """
        if options is None:
            options = {}

        if isinstance(options, Options):
            client_options = options.get_struct()
        else:
            client_options = options

        with open(metalink_file_path, "rb") as stream:
            metalink_contents = stream.read()
        encoded_contents = b64encode(metalink_contents).decode("utf8")

        gids = self.client.add_metalink(encoded_contents, client_options, position)

        return self.get_downloads(gids)

    def add_uris(
        self,
        uris: List[str],
        options: Optional[OptionsType] = None,
        position: Optional[int] = None,
    ) -> Download:
        """
        Add a download with a URL (or more).

        Arguments:
            uris: A list of URIs that point to the same resource.
            options: An instance of the `Options` class or a dictionary
                containing aria2c options to create the download with.
            position: The position where to insert the new download in the queue. Start at 0 (top).

        Returns:
            The newly created download object.

        """
        if options is None:
            options = {}

        if isinstance(options, Options):
            client_options = options.get_struct()
        else:
            client_options = options

        gid = self.client.add_uri(uris, client_options, position)

        return self.get_download(gid)

    def search(self, patterns: List[str]) -> List[Download]:
        """
        Not implemented.

        Search and return [`Download`][aria2p.downloads.Download] objects based on multiple patterns.

        Arguments:
            patterns: The patterns used to filter the download list.

        Raises:
            NotImplementedError: This method is not implemented yet.
        """
        # gid
        # status
        # totalLength
        # completedLength
        # uploadLength
        # bitfield
        # downloadSpeed
        # uploadSpeed
        # infoHash
        # numSeeders
        # seeder
        # pieceLength
        # numPieces
        # connections
        # errorCode
        # errorMessage
        # followedBy
        # following
        # belongsTo
        # dir
        # files
        # bittorrent
        #        announceList
        #        comment
        #        creationDate
        #        mode
        #        info
        #               name
        # verifiedLength
        # verifyIntegrityPending
        raise NotImplementedError

    def get_download(self, gid: str) -> Download:
        """
        Get a [`Download`][aria2p.downloads.Download] object thanks to its GID.

        Arguments:
            gid: The GID of the download to get.

        Returns:
            The retrieved download object.
        """
        return Download(self, self.client.tell_status(gid))

    def get_downloads(self, gids: List[str] = None) -> List[Download]:
        """
        Get a list of [`Download`][aria2p.downloads.Download] object thanks to their GIDs.

        Arguments:
            gids: The GIDs of the downloads to get. If None, return all the downloads.

        Returns:
            The retrieved download objects.
        """
        downloads = []

        if gids:
            for gid in gids:
                downloads.append(Download(self, self.client.tell_status(gid)))
        else:
            structs = []
            structs.extend(self.client.tell_active())
            structs.extend(self.client.tell_waiting(0, 1000))
            structs.extend(self.client.tell_stopped(0, 1000))
            downloads = [Download(self, struct) for struct in structs]

        return downloads

    def move(self, download: Download, pos: int) -> int:
        """
        Move a download in the queue, relatively to its current position.

        Arguments:
            download: The download object to move.
            pos: The relative position (1 to move down, -1 to move up, -2 to move up two times, etc.).

        Returns:
            The new position of the download.
        """
        return self.client.change_position(download.gid, pos, "POS_CUR")

    def move_to(self, download: Download, pos: int) -> int:
        """
        Move a download in the queue, with absolute positioning.

        Arguments:
            download: The download object to move.
            pos: The absolute position in the queue where to move the download. 0 for top, -1 for bottom.

        Returns:
            The new position of the download.
        """
        if pos < 0:
            how = "POS_END"
            pos = -pos
        else:
            how = "POS_SET"
        return self.client.change_position(download.gid, pos, how)

    def move_up(self, download: Download, pos: int = 1) -> int:
        """
        Move a download up in the queue.

        Arguments:
            download: The download object to move.
            pos: Number of times to move up. With negative values, will move down (use move or move_down instead).

        Returns:
            The new position of the download.
        """
        return self.client.change_position(download.gid, -pos, "POS_CUR")

    def move_down(self, download: Download, pos: int = 1) -> int:
        """
        Move a download down in the queue.

        Arguments:
            download: The download object to move.
            pos: Number of times to move down. With negative values, will move up (use move or move_up instead).

        Returns:
            The new position of the download.
        """
        return self.client.change_position(download.gid, pos, "POS_CUR")

    def move_to_top(self, download: Download) -> int:
        """
        Move a download to the top of the queue.

        Arguments:
            download: The download object to move.

        Returns:
            The new position of the download.
        """
        return self.client.change_position(download.gid, 0, "POS_SET")

    def move_to_bottom(self, download: Download) -> int:
        """
        Move a download to the bottom of the queue.

        Arguments:
            download: The download object to move.

        Returns:
            The new position of the download.
        """
        return self.client.change_position(download.gid, 0, "POS_END")

    def retry_downloads(  # noqa: WPS231 (not that complex)
        self,
        downloads: List[Download],
        clean: bool = False,
    ) -> List[OperationResult]:
        """
        Resume failed downloads from where they left off with new GIDs.

        Arguments:
            downloads: The list of downloads to remove.
            clean: Whether to remove the aria2 control file as well.

        Returns:
            Success or failure of the operation for each given download.
        """
        result: List[OperationResult] = []

        for download in downloads:
            if not download.has_failed:
                continue
            uri = download.files[0].uris[0]["uri"]  # noqa: WPS219 (deep access)
            try:
                new_download_gid = self.add_uris([uri], download.options)
            except ClientException as error:
                result.append(error)
            else:
                if not new_download_gid:
                    continue

                self.remove([download], clean)
                result.append(True)

        return result

    def remove(  # noqa: WPS231 (complex, maybe we could split it)
        self,
        downloads: List[Download],
        force: bool = False,
        files: bool = False,
        clean: bool = True,
    ) -> List[OperationResult]:
        """
        Remove the given downloads from the list.

        Arguments:
            downloads: The list of downloads to remove.
            force: Whether to force the removal or not.
            files: Whether to remove downloads files as well.
            clean: Whether to remove the aria2 control file as well.

        Returns:
            Success or failure of the operation for each given download.
        """
        # Note: batch/multicall candidate
        if force:
            remove_func = self.client.force_remove
        else:
            remove_func = self.client.remove

        result: List[OperationResult] = []

        for download in downloads:
            if download.is_complete or download.is_removed or download.has_failed:
                logger.debug(f"Try to remove download result {download.gid}")
                try:
                    self.client.remove_download_result(download.gid)
                except ClientException as error:
                    logger.exception(error)
                    result.append(error)
                else:
                    logger.success(f"Removed download result {download.gid}")
                    result.append(True)
            else:
                logger.debug(f"Try to remove download {download.gid}")
                try:
                    removed_gid = remove_func(download.gid)
                except ClientException as error:  # noqa: WPS440 (block variable overlap)
                    logger.exception(error)
                    result.append(error)
                else:
                    logger.success(f"Removed download {download.gid}")
                    result.append(True)
                    try:  # noqa: WPS505 (nested try)
                        self.client.remove_download_result(download.gid)
                    except ClientException as error:  # noqa: WPS440 (block variable overlap)
                        logger.debug(f"Failed to remove download result {download.gid}")
                        logger.opt(exception=True).trace(error)
                    if removed_gid != download.gid:
                        logger.debug(
                            f"Removed download GID#{removed_gid} is different than download GID#{download.gid}",
                        )
                        try:  # noqa: WPS505 (nested try)
                            self.client.remove_download_result(removed_gid)
                        except ClientException as error:  # noqa: WPS440 (block variable overlap)
                            logger.debug(f"Failed to remove download result {removed_gid}")
                            logger.opt(exception=True).trace(error)

            if clean:
                # FUTURE: use missing_ok parameter on Python 3.8
                try:
                    download.control_file_path.unlink()
                except FileNotFoundError:
                    logger.debug(f"aria2 control file {download.control_file_path} was not found")
                else:
                    logger.debug(f"Removed control file {download.control_file_path}")

            if files and result[-1]:
                self.remove_files([download], force=True)

        return result

    def remove_all(self, force: bool = False) -> bool:
        """
        Remove all downloads from the list.

        Arguments:
            force: Whether to force the removal or not.

        Returns:
            Success or failure of the operation to remove all downloads.
        """
        return all(self.remove(self.get_downloads(), force=force))

    def pause(self, downloads: List[Download], force: bool = False) -> List[OperationResult]:
        """
        Remove the given downloads from the list.

        Arguments:
            downloads: The list of downloads to remove.
            force: Whether to pause immediately without contacting servers or not.

        Returns:
            Success or failure of the operation for each given download.
        """
        # Note: batch/multicall candidate
        if force:
            pause_func = self.client.force_pause
        else:
            pause_func = self.client.pause

        result: List[OperationResult] = []

        for download in downloads:
            try:
                pause_func(download.gid)
            except ClientException as error:
                logger.debug(f"Failed to pause download {download.gid}")
                logger.opt(exception=True).trace(error)
                result.append(error)
            else:
                result.append(True)

        return result

    def pause_all(self, force: bool = False) -> bool:
        """
        Remove the given downloads from the list.

        Arguments:
            force: Whether to pause immediately without contacting servers or not.

        Returns:
            Success or failure of the operation to pause all downloads.
        """
        if force:
            pause_func = self.client.force_pause_all
        else:
            pause_func = self.client.pause_all
        return pause_func() == "OK"

    def resume(self, downloads: List[Download]) -> List[OperationResult]:
        """
        Resume (unpause) the given downloads.

        Arguments:
            downloads: The list of downloads to resume.

        Returns:
            Success or failure of the operation for each given download.
        """
        # Note: batch/multicall candidate
        result: List[OperationResult] = []

        for download in downloads:
            try:
                self.client.unpause(download.gid)
            except ClientException as error:
                logger.debug(f"Failed to resume download {download.gid}")
                logger.opt(exception=True).trace(error)
                result.append(error)
            else:
                result.append(True)

        return result

    def resume_all(self) -> bool:
        """
        Resume (unpause) all downloads.

        Returns:
            Success or failure of the operation to resume all downloads.
        """
        return self.client.unpause_all() == "OK"

    def purge(self) -> bool:
        """
        Purge completed, removed or failed downloads from the queue.

        Returns:
            Success or failure of the operation.
        """
        return self.client.purge_download_result() == "OK"

    def autopurge(self) -> bool:
        """
        Purge completed, removed or failed downloads from the queue.

        Deprecated. Use [`purge`][aria2p.api.API.purge] instead.

        Returns:
            Success or failure of the operation.
        """
        logger.warning("Deprecation warning: API method 'autopurge' is deprecated, use 'purge' instead.")
        return self.purge()

    def get_options(self, downloads: List[Download]) -> List[Options]:
        """
        Get options for each of the given downloads.

        Arguments:
            downloads: The list of downloads to get the options of.

        Returns:
            Options object for each given download.
        """
        # Note: batch/multicall candidate
        options = []
        for download in downloads:
            options.append(Options(self, self.client.get_option(download.gid), download))
        return options

    def get_global_options(self) -> Options:
        """
        Get the global options.

        Returns:
            The global aria2c options.
        """
        return Options(self, self.client.get_global_option())

    def set_options(self, options: OptionsType, downloads: List[Download]) -> List[bool]:
        """
        Set options for specific downloads.

        Arguments:
            options: An instance of the [`Options`][aria2p.options.Options] class or a dictionary
                containing aria2c options to create the download with.
            downloads: The list of downloads to set the options for.

        Returns:
            Success or failure of the operation for changing options for each given download.
        """
        if isinstance(options, Options):
            client_options = options.get_struct()
        else:
            client_options = options

        # Note: batch/multicall candidate
        results = []
        for download in downloads:
            results.append(self.client.change_option(download.gid, client_options) == "OK")
        return results

    def set_global_options(self, options: OptionsType) -> bool:
        """
        Set global options.

        Arguments:
            options: An instance of the [`Options`][aria2p.options.Options] class or a dictionary
                containing aria2c options to create the download with.

        Returns:
            Success or failure of the operation for changing global options.
        """
        if isinstance(options, Options):
            client_options = options.get_struct()
        else:
            client_options = options

        return self.client.change_global_option(client_options) == "OK"

    def get_stats(self) -> Stats:
        """
        Get the stats of the remote aria2c process.

        Returns:
            The global stats returned by the remote process.
        """
        return Stats(self.client.get_global_stat())

    @staticmethod  # noqa: WPS231,WPS602 (complex method, static method)
    def remove_files(  # noqa: WPS231,WPS602
        downloads: List[Download],
        force: bool = False,
    ) -> List[bool]:
        """
        Remove downloaded files.

        Arguments:
            downloads:  the list of downloads for which to remove files.
            force: Whether to remove files even if download is not complete.

        Returns:
            Success or failure of the operation for each given download.
        """
        results = []
        for download in downloads:
            if download.is_complete or force:
                for path in download.root_files_paths:
                    if path.is_dir():
                        try:
                            shutil.rmtree(str(path))
                        except OSError:
                            logger.error(f"Could not delete directory '{path}'")
                            logger.opt(exception=True).trace(error)
                            results.append(False)
                        else:
                            results.append(True)
                    else:
                        try:
                            path.unlink()
                        except FileNotFoundError as error:
                            logger.warning(f"File '{path}' did not exist when trying to delete it")
                            logger.opt(exception=True).trace(error)
                        results.append(True)
            else:
                results.append(False)
        return results

    @staticmethod  # noqa: WPS602 (static method)
    def move_files(  # noqa: WPS602
        downloads: List[Download],
        to_directory: PathOrStr,
        force: bool = False,
    ) -> List[bool]:
        """
        Move downloaded files to another directory.

        Arguments:
            downloads:  the list of downloads for which to move files.
            to_directory: The target directory to move files to.
            force: Whether to move files even if download is not complete.

        Returns:
            Success or failure of the operation for each given download.
        """
        if isinstance(to_directory, str):
            to_directory = Path(to_directory)

        # raises FileExistsError when target is already a file
        to_directory.mkdir(parents=True, exist_ok=True)

        results = []
        for download in downloads:
            if download.is_complete or force:
                for path in download.root_files_paths:
                    shutil.move(str(path), str(to_directory))
                results.append(True)
            else:
                results.append(False)
        return results

    @staticmethod  # noqa: WPS231,WPS602 (not that complex, static method)
    def copy_files(  # noqa: WPS231,WPS602 (not that complex)
        downloads: List[Download],
        to_directory: PathOrStr,
        force: bool = False,
    ) -> List[bool]:
        """
        Copy downloaded files to another directory.

        Arguments:
            downloads:  the list of downloads for which to move files.
            to_directory: The target directory to copy files into.
            force: Whether to move files even if download is not complete.

        Returns:
            Success or failure of the operation for each given download.
        """
        if isinstance(to_directory, str):
            to_directory = Path(to_directory)

        # raises FileExistsError when target is already a file
        to_directory.mkdir(parents=True, exist_ok=True)

        results = []
        for download in downloads:
            if download.is_complete or force:
                for path in download.root_files_paths:
                    if path.is_dir():
                        shutil.copytree(str(path), str(to_directory / path.name))
                    elif path.is_file():
                        shutil.copy(str(path), str(to_directory))

                results.append(True)
            else:
                results.append(False)
        return results

    def listen_to_notifications(
        self,
        threaded: bool = False,
        on_download_start: Callable = None,
        on_download_pause: Callable = None,
        on_download_stop: Callable = None,
        on_download_complete: Callable = None,
        on_download_error: Callable = None,
        on_bt_download_complete: Callable = None,
        timeout: int = 5,
        handle_signals: bool = True,
    ) -> None:
        """
        Start listening to aria2 notifications via WebSocket.

        This method differs from [`Client.listen_to_notifications`][aria2p.client.Client.listen_to_notifications]
        in that it expects callbacks accepting two arguments, `api` and `gid`, instead of only `gid`.
        Accepting `api` allows to use the high-level methods of the [`API`][aria2p.api.API] class.

        Stop listening to notifications with the [`API.stop_listening`][aria2p.api.API.stop_listening] method.

        Arguments:
            threaded: Whether to start the listening loop in a thread or not (non-blocking or blocking).
            on_download_start: Callback for the `onDownloadStart` event.
            on_download_pause: Callback for the `onDownloadPause` event.
            on_download_stop: Callback for the `onDownloadStop` event.
            on_download_complete: Callback for the `onDownloadComplete` event.
            on_download_error: Callback for the `onDownloadError` event.
            on_bt_download_complete: Callback for the `onBtDownloadComplete` event.
            timeout: Timeout when waiting for data to be received. Use a small value for faster reactivity
                when stopping to listen. Default is 5 seconds.
            handle_signals: Whether to add signal handlers to gracefully stop the loop on SIGTERM and SIGINT.
        """

        def closure(callback):  # noqa: WPS430 (nested function)
            return functools.partial(callback, self) if callable(callback) else None

        kwargs = {
            "on_download_start": closure(on_download_start),
            "on_download_pause": closure(on_download_pause),
            "on_download_stop": closure(on_download_stop),
            "on_download_complete": closure(on_download_complete),
            "on_download_error": closure(on_download_error),
            "on_bt_download_complete": closure(on_bt_download_complete),
            "timeout": timeout,
            "handle_signals": handle_signals,
        }

        if threaded:
            kwargs["handle_signals"] = False
            self.listener = threading.Thread(target=self.client.listen_to_notifications, kwargs=kwargs)
            self.listener.start()
        else:
            self.client.listen_to_notifications(**kwargs)

    def stop_listening(self) -> None:
        """
        Stop listening to notifications.

        If the listening loop was threaded, this method will wait for the thread to finish.
        The time it takes for the thread to finish will depend on the timeout given while calling
        [`listen_to_notifications`][aria2p.api.API.listen_to_notifications].
        """
        self.client.stop_listening()
        if self.listener:
            self.listener.join()
        self.listener = None



"""
This module defines the BitTorrent, File and Download classes.

They respectively hold structured information about
torrent files, files and downloads in aria2c.
"""

from datetime import datetime, timedelta
from pathlib import Path
from typing import List, Optional

from loguru import logger

import aria2p
from aria2p.client import ClientException
from aria2p.options import Options
from aria2p.types import PathOrStr
from aria2p.utils import bool_or_value, human_readable_bytes, human_readable_timedelta


class BitTorrent:
    """Information retrieved from a torrent file."""

    def __init__(self, struct: dict) -> None:
        """
        Initialize the object.

        Arguments:
            struct: A dictionary Python object returned by the JSON-RPC client.
        """
        self._struct = struct or {}

    def __str__(self):
        return self.info["name"]

    @property  # noqa: WPS234 (not complex type annotation)
    def announce_list(self) -> Optional[List[List[str]]]:  # noqa: WPS234
        """
        List of lists of announce URIs.

        If the torrent contains announce and no announce-list, announce is converted to the announce-list format.

        Returns:
            The announce URIs.
        """
        return self._struct.get("announceList")

    @property
    def comment(self) -> Optional[str]:
        """
        Return the comment of the torrent.

        comment.utf-8 is used if available.

        Returns:
            The torrent's comment.
        """
        return self._struct.get("comment")

    @property
    def creation_date(self) -> datetime:
        """
        Return the creation time of the torrent.

        The value is an integer since the epoch, measured in seconds.

        Returns:
            The creation date.
        """
        return datetime.fromtimestamp(self._struct["creationDate"])

    @property
    def mode(self) -> Optional[str]:
        """
        File mode of the torrent.

        The value is either single or multi.

        Returns:
            The file mode.
        """
        return self._struct.get("mode")

    @property
    def info(self) -> Optional[dict]:
        """
        Struct which contains data from Info dictionary.

        It contains the `name` key: name in info dictionary. `name.utf-8` is used if available.

        Returns:
            The torrent's info.
        """
        return self._struct.get("info")


class File:
    """Information about a download's file."""

    def __init__(self, struct: dict) -> None:
        """
        Initialize the object.

        Arguments:
            struct: A dictionary Python object returned by the JSON-RPC client.
        """
        self._struct = struct or {}

    def __str__(self):
        return str(self.path)

    def __eq__(self, other):
        return self.path == other.path

    @property
    def index(self) -> int:
        """
        Index of the file, starting at 1, in the same order as files appear in the multi-file torrent.

        Returns:
            The index of the file.
        """
        return int(self._struct["index"])

    @property
    def path(self) -> Path:
        """
        File path.

        Returns:
            The file path.
        """
        return Path(self._struct["path"])

    @property
    def is_metadata(self) -> bool:
        """
        Return True if this file is aria2 metadata and not an actual file.

        Returns:
            If the file is metadata.
        """
        return str(self.path).startswith("[METADATA]")

    @property
    def length(self) -> int:
        """
        Return the file size in bytes.

        Returns:
            The file size in bytes.
        """
        return int(self._struct["length"])

    def length_string(self, human_readable: bool = True) -> str:
        """
        Return the length as string.

        Arguments:
            human_readable: Return in human readable format or not.

        Returns:
            The length string.
        """
        if human_readable:
            return human_readable_bytes(self.length, delim=" ")
        return str(self.length) + " B"

    @property
    def completed_length(self) -> int:
        """
        Completed length of this file in bytes.

        Please note that it is possible that sum of completedLength is less than the completedLength returned by the
        aria2.tellStatus() method. This is because completedLength in aria2.getFiles() only includes completed
        pieces. On the other hand, completedLength in aria2.tellStatus() also includes partially completed pieces.

        Returns:
            The completed length.
        """
        return int(self._struct["completedLength"])

    def completed_length_string(self, human_readable: bool = True) -> str:
        """
        Return the completed length as string.

        Arguments:
            human_readable: Return in human readable format or not.

        Returns:
            The completed length string.
        """
        if human_readable:
            return human_readable_bytes(self.completed_length, delim=" ")
        return str(self.completed_length) + " B"

    @property
    def selected(self) -> bool:
        """
        Return True if this file is selected by [`--select-file`][aria2p.options.Options.select_file] option.

        If [`--select-file`][aria2p.options.Options.select_file] is not specified
        or this is single-file torrent or not a torrent download at all, this value is always true.
        Otherwise false.

        Returns:
            If this file is selected.
        """
        return bool_or_value(self._struct["selected"])

    @property
    def uris(self) -> Optional[List[str]]:
        """
        Return a list of URIs for this file.

        The element type is the same struct
        used in the [`client.get_uris()`][aria2p.client.Client.get_uris] method.

        Returns:
            The list of URIs.
        """
        return self._struct.get("uris")


class Download:
    """Class containing all information about a download, as retrieved with the client."""

    def __init__(self, api: "aria2p.api.API", struct: dict) -> None:
        """
        Initialize the object.

        Arguments:
            api: The reference to an [`API`][aria2p.api.API] instance.
            struct: A dictionary Python object returned by the JSON-RPC client.
        """
        self.api = api
        self._struct = struct or {}
        self._files: List[File] = []
        self._root_files_paths: List[Path] = []
        self._bittorrent = None
        self._name = ""
        self._options: Optional[Options] = None
        self._followed_by: Optional[List[Download]] = None
        self._following: Optional[Download] = None
        self._belongs_to: Optional[Download] = None

    def __str__(self):
        return self.name

    def __eq__(self, other):
        return self.gid == other.gid

    def update(self) -> None:
        """Update the internal values of the download with more recent values."""
        self._struct = self.api.client.tell_status(self.gid)

        self._files = []
        self._name = ""
        self._bittorrent = None
        self._followed_by = None
        self._following = None
        self._belongs_to = None
        self._options = None

    @property
    def live(self) -> "Download":
        """
        Return the same object with updated data.

        Returns:
            Itself.
        """
        self.update()
        return self

    @property  # noqa: WPS231 (not that complex)
    def name(self) -> str:  # noqa: WPS231
        """
        Return the name of the download.

        Name is the name of the file if single-file, first file's directory name if multi-file.

        Returns:
            The download name.
        """
        if not self._name:
            if self.bittorrent and self.bittorrent.info:
                self._name = self.bittorrent.info["name"]
            elif self.files[0].is_metadata:
                self._name = str(self.files[0].path)
            else:
                file_path = str(self.files[0].path.absolute())
                dir_path = str(self.dir.absolute())
                if file_path.startswith(dir_path):
                    start_pos = len(dir_path) + 1
                    self._name = Path(file_path[start_pos:]).parts[0]
                else:
                    try:
                        self._name = self.files[0].uris[0]["uri"].split("/")[-1]  # noqa: WPS221 (not complex)
                    except IndexError:
                        pass  # noqa: WPS420 (we don't want to fail here)
        return self._name

    @property
    def control_file_path(self) -> Path:
        """
        Return the path to the aria2 control file for this download.

        Returns:
            The control file path.
        """
        return self.dir / (self.name + ".aria2")

    @property  # noqa: WPS231 (not that complex)
    def root_files_paths(self) -> List[Path]:  # noqa: WPS231
        """
        Return the unique set of directories/files for this download.

        Instead of returning all the leaves like self.files,
        return the relative root directories if any, and relative root files.

        This property is useful when we need to list the directories and files
        in order to move or copy them. We don't want to copy files one by one,
        but rather entire directories at once when possible.

        Returns:
            The root file paths.

        Examples:
            Download directory is `/a/b`.

            >>> self.files
            ["/a/b/c/1.txt", "/a/b/c/2.txt", "/a/b/3.txt"]
            >>> self.root_files_paths
            ["/a/b/c", "/a/b/3.txt"]
        """
        if not self._root_files_paths:
            paths = []
            for file in self.files:  # noqa: VNE002 (file)
                if file.is_metadata:
                    continue
                try:
                    relative_path = file.path.relative_to(self.dir)
                except ValueError as error:
                    logger.warning(f"Can't determine file path '{file.path}' relative to '{self.dir}'")
                    logger.opt(exception=True).trace(error)
                else:
                    path = self.dir / relative_path.parts[0]
                    if path not in paths:
                        paths.append(path)
            self._root_files_paths = paths
        return self._root_files_paths

    @property
    def options(self) -> Options:
        """
        Options specific to this download.

        Returns:
            The download options.
        """
        if not self._options:
            self.update_options()
        return self._options  # type: ignore

    @options.setter
    def options(self, value):
        self._options = value

    def update_options(self) -> None:
        """Re-fetch the options from the remote."""
        self._options = self.api.get_options(downloads=[self])[0]

    @property
    def gid(self) -> str:
        """
        GID of the download.

        Returns:
            The download GID.
        """
        return self._struct["gid"]

    @property
    def status(self) -> str:
        """
        Return the status of the download.

        Returns:
            `active`, `waiting`, `paused`, `error`, `complete` or `removed`.
        """
        return self._struct["status"]

    @property
    def is_active(self) -> bool:
        """
        Return True if download is active.

        Returns:
            If this download is active.
        """
        return self.status == "active"

    @property
    def is_waiting(self) -> bool:
        """
        Return True if download is waiting.

        Returns:
            If this download is waiting.
        """
        return self.status == "waiting"

    @property
    def is_paused(self) -> bool:
        """
        Return True if download is paused.

        Returns:
            If this download is paused.
        """
        return self.status == "paused"

    @property
    def has_failed(self) -> bool:
        """
        Return True if download has errored.

        Returns:
            If this download has failed.
        """
        return self.status == "error"

    @property
    def is_complete(self) -> bool:
        """
        Return True if download is complete.

        Returns:
            If this download is complete.
        """
        return self.status == "complete"

    @property
    def is_removed(self) -> bool:
        """
        Return True if download was removed.

        Returns:
            If this download was removed.
        """
        return self.status == "removed"

    @property
    def is_metadata(self) -> bool:
        """
        Return True if this download is only composed of metadata, and no actual files.

        Returns:
            If this is a metadata download.
        """
        return all(_.is_metadata for _ in self.files)

    @property
    def is_torrent(self) -> bool:
        """
        Return true if this download is a torrent.

        Returns:
            If this is a torrent downlaod.
        """
        return "bittorrent" in self._struct

    @property
    def total_length(self) -> int:
        """
        Total length of the download in bytes.

        Returns:
            The total length in bytes.
        """
        return int(self._struct["totalLength"])

    def total_length_string(self, human_readable: bool = True) -> str:
        """
        Return the total length as string.

        Arguments:
            human_readable: Return in human readable format or not.

        Returns:
            The total length string.
        """
        if human_readable:
            return human_readable_bytes(self.total_length, delim=" ")
        return str(self.total_length) + " B"

    @property
    def completed_length(self) -> int:
        """
        Completed length of the download in bytes.

        Returns:
            The completed length in bytes.
        """
        return int(self._struct["completedLength"])

    def completed_length_string(self, human_readable: bool = True) -> str:
        """
        Return the completed length as string.

        Arguments:
            human_readable: Return in human readable format or not.

        Returns:
            The completed length string.
        """
        if human_readable:
            return human_readable_bytes(self.completed_length, delim=" ")
        return str(self.completed_length) + " B"

    @property
    def upload_length(self) -> int:
        """
        Return the uploaded length of the download in bytes.

        Returns:
            The uploaded length in bytes.
        """
        return int(self._struct["uploadLength"])

    def upload_length_string(self, human_readable: bool = True) -> str:
        """
        Return the upload length as string.

        Arguments:
            human_readable: Return in human readable format or not.

        Returns:
            The upload length string.
        """
        if human_readable:
            return human_readable_bytes(self.upload_length, delim=" ")
        return str(self.upload_length) + " B"

    @property
    def bitfield(self) -> Optional[str]:
        """
        Hexadecimal representation of the download progress.

        The highest bit corresponds to the piece at index 0. Any set bits indicate loaded pieces, while unset bits
        indicate not yet loaded and/or missing pieces. Any overflow bits at the end are set to zero. When the
        download was not started yet, this key will not be included in the response.

        Returns:
            The hexadecimal representation of the download progress.
        """
        return self._struct.get("bitfield")

    @property
    def download_speed(self) -> int:
        """
        Download speed of this download measured in bytes/sec.

        Returns:
            The download speed in bytes/sec.
        """
        return int(self._struct["downloadSpeed"])

    def download_speed_string(self, human_readable: bool = True) -> str:
        """
        Return the download speed as string.

        Arguments:
            human_readable: Return in human readable format or not.

        Returns:
            The download speed string.
        """
        if human_readable:
            return human_readable_bytes(self.download_speed, delim=" ", postfix="/s")
        return str(self.download_speed) + " B/s"

    @property
    def upload_speed(self) -> int:
        """
        Upload speed of this download measured in bytes/sec.

        Returns:
            The upload speed in bytes/sec.
        """
        return int(self._struct["uploadSpeed"])

    def upload_speed_string(self, human_readable: bool = True) -> str:
        """
        Return the upload speed as string.

        Arguments:
            human_readable: Return in human readable format or not.

        Returns:
            The upload speed string.
        """
        if human_readable:
            return human_readable_bytes(self.upload_speed, delim=" ", postfix="/s")
        return str(self.upload_speed) + " B/s"

    @property
    def info_hash(self) -> Optional[str]:
        """
        Return the InfoHash.

        BitTorrent only.

        Returns:
            The InfoHash.
        """
        return self._struct.get("infoHash")

    @property
    def num_seeders(self) -> int:
        """
        Return the number of seeders aria2 has connected to.

        BitTorrent only.

        Returns:
            The numbers of seeders.
        """
        return int(self._struct["numSeeders"])

    @property
    def seeder(self) -> bool:
        """
        Return True if the local endpoint is a seeder, otherwise false.

        BitTorrent only.

        Returns:
            If the local endpoint is a seeder.
        """
        return bool_or_value(self._struct.get("seeder"))

    @property
    def piece_length(self) -> int:
        """
        Piece length in bytes.

        Returns:
            The piece length in bytes.
        """
        return int(self._struct["pieceLength"])

    def piece_length_string(self, human_readable: bool = True) -> str:
        """
        Return the piece length as string.

        Arguments:
            human_readable: Return in human readable format or not.

        Returns:
            The piece length string.
        """
        if human_readable:
            return human_readable_bytes(self.piece_length, delim=" ")
        return str(self.piece_length) + " B"

    @property
    def num_pieces(self) -> int:
        """
        Return the number of pieces.

        Returns:
            The number of pieces.
        """
        return int(self._struct["numPieces"])

    @property
    def connections(self) -> int:
        """
        Return the number of peers/servers aria2 has connected to.

        Returns:
            The number of connected peers/servers.
        """
        return int(self._struct["connections"])

    @property
    def error_code(self) -> Optional[str]:
        """
        Return the code of the last error for this item, if any.

        The value is a string. The error codes are defined in the EXIT STATUS section. This value is only available
        for stopped/completed downloads.

        Returns:
            The error code.
        """
        return self._struct.get("errorCode")

    @property
    def error_message(self) -> Optional[str]:
        """
        Return the (hopefully) human readable error message associated to errorCode.

        Returns:
            The error message.
        """
        return self._struct.get("errorMessage")

    @property
    def followed_by_ids(self) -> List[str]:
        """
        List of GIDs which are generated as the result of this download.

        For example, when aria2 downloads a Metalink file, it generates downloads described in the Metalink (see the
        --follow-metalink option). This value is useful to track auto-generated downloads. If there are no such
        downloads, this key will not be included in the response.

        Returns:
            The children downloads IDs.
        """
        return self._struct.get("followedBy", [])

    @property
    def followed_by(self) -> List["Download"]:
        """
        List of downloads generated as the result of this download.

        Returns:
            A list of instances of [`Download`][aria2p.downloads.Download].
        """
        if self._followed_by is None:
            result = []
            for gid in self.followed_by_ids:
                try:
                    result.append(self.api.get_download(gid))
                except ClientException as error:
                    logger.warning(
                        f"Can't find download with GID {gid}, try to update download {self.gid} ({id(self)}",  # noqa: WPS221
                    )
                    logger.opt(exception=True).trace(error)
            self._followed_by = result
        return self._followed_by

    @property
    def following_id(self) -> Optional[str]:
        """
        Return the reverse link for followedBy.

        A download included in followedBy has this object's GID in its following value.

        Returns:
            The parent download ID.
        """
        return self._struct.get("following")

    @property
    def following(self) -> Optional["Download"]:
        """
        Return the download this download is following.

        Returns:
            An instance of [`Download`][aria2p.downloads.Download].
        """
        if not self._following:
            following_id = self.following_id
            if following_id:
                try:
                    self._following = self.api.get_download(following_id)
                except ClientException as error:
                    logger.warning(
                        f"Can't find download with GID {following_id}, try to update download {self.gid} ({id(self)}",
                    )
                    logger.opt(exception=True).trace(error)
                    self._following = None
        return self._following

    @property
    def belongs_to_id(self) -> Optional[str]:
        """
        GID of a parent download.

        Some downloads are a part of another download. For example, if a file in a Metalink has BitTorrent resources,
        The downloads of ".torrent" files are parts of that parent. If this download has no parent, this key will not
        be included in the response.

        Returns:
            The GID of the parent download.
        """
        return self._struct.get("belongsTo")

    @property
    def belongs_to(self) -> Optional["Download"]:
        """
        Parent download.

        Returns:
            An instance of [`Download`][aria2p.downloads.Download].
        """
        if not self._belongs_to:
            belongs_to_id = self.belongs_to_id
            if belongs_to_id:
                try:
                    self._belongs_to = self.api.get_download(belongs_to_id)
                except ClientException as error:
                    logger.warning(
                        f"Can't find download with GID {belongs_to_id}, try to update download {self.gid} ({id(self)})",
                    )
                    logger.opt(exception=True).trace(error)
                    self._belongs_to = None
        return self._belongs_to

    @property  # noqa: A003 (shadowing dir)
    def dir(self) -> Path:  # noqa: A003
        """
        Directory to save files.

        Returns:
            The directory where the files are saved.
        """
        return Path(self._struct["dir"])

    @property
    def files(self) -> List[File]:
        """
        Return the list of files.

        The elements of this list are the same structs used in aria2.getFiles() method.

        Returns:
            The files of this download.
        """
        if not self._files:
            self._files = [File(struct) for struct in self._struct.get("files", [])]
        return self._files

    @property
    def bittorrent(self) -> Optional[BitTorrent]:
        """
        Struct which contains information retrieved from the .torrent (file).

        BitTorrent only.

        Returns:
            A [BitTorrent][aria2p.downloads.BitTorrent] instance or `None`.
        """
        if not self._bittorrent and "bittorrent" in self._struct:
            self._bittorrent = BitTorrent(self._struct.get("bittorrent"))
        return self._bittorrent

    @property
    def verified_length(self) -> int:
        """
        Return the number of verified number of bytes while the files are being hash checked.

        This key exists only when this download is being hash checked.

        Returns:
            The verified length.
        """
        return int(self._struct.get("verifiedLength", 0))

    def verified_length_string(self, human_readable: bool = True) -> str:
        """
        Return the verified length as string.

        Arguments:
            human_readable: Return in human readable format or not.

        Returns:
            The verified length string.
        """
        if human_readable:
            return human_readable_bytes(self.verified_length, delim=" ")
        return str(self.verified_length) + " B"

    @property
    def verify_integrity_pending(self) -> Optional[bool]:
        """
        Return True if this download is waiting for the hash check in a queue.

        This key exists only when this download is in the queue.

        Returns:
            Whether this download is waiting for the hash check.
        """
        return bool_or_value(self._struct.get("verifyIntegrityPending"))

    @property
    def progress(self) -> float:
        """
        Return the progress of the download as float.

        Returns:
            Progress percentage.
        """
        try:
            return self.completed_length / self.total_length * 100
        except ZeroDivisionError:
            return 0.0

    def progress_string(self, digits: int = 2) -> str:
        """
        Return the progress percentage as string.

        Arguments:
            digits: Number of decimal digits to use.

        Returns:
            The progress percentage.
        """
        return f"{self.progress:.{digits}f}%"

    @property
    def eta(self) -> timedelta:
        """
        Return the Estimated Time of Arrival (a timedelta).

        Returns:
            ETA or `timedelta.max` if unknown.
        """
        try:
            return timedelta(seconds=int((self.total_length - self.completed_length) / self.download_speed))
        except ZeroDivisionError:
            return timedelta.max

    def eta_string(self, precision: int = 0) -> str:
        """
        Return the Estimated Time of Arrival as a string.

        Arguments:
            precision: The precision to use, see [aria2p.utils.human_readable_timedelta].

        Returns:
            The Estimated Time of Arrival as a string.
        """
        eta = self.eta

        if eta == timedelta.max:
            return "-"

        return human_readable_timedelta(eta, precision=precision)

    def move(self, pos: int) -> int:
        """
        Move the download in the queue, relatively.

        Arguments:
            pos: Number of times to move.

        Returns:
            The new position of the download.
        """
        return self.api.move(self, pos)

    def move_to(self, pos: int) -> int:
        """
        Move the download in the queue, absolutely.

        Arguments:
            pos: The absolute position in the queue to take.

        Returns:
            The new position of the download.
        """
        return self.api.move_to(self, pos)

    def move_up(self, pos: int = 1) -> int:
        """
        Move the download up in the queue.

        Arguments:
            pos: Number of times to move up.

        Returns:
            The new position of the download.
        """
        return self.api.move_up(self, pos)

    def move_down(self, pos: int = 1) -> int:
        """
        Move the download down in the queue.

        Arguments:
            pos: Number of times to move down.

        Returns:
            The new position of the download.
        """
        return self.api.move_down(self, pos)

    def move_to_top(self) -> int:
        """
        Move the download to the top of the queue.

        Returns:
            The new position of the download.
        """
        return self.api.move_to_top(self)

    def move_to_bottom(self) -> int:
        """
        Move the download to the bottom of the queue.

        Returns:
            The new position of the download.
        """
        return self.api.move_to_bottom(self)

    def remove(self, force: bool = False, files: bool = False) -> bool:
        """
        Remove the download from the queue (even if active).

        Arguments:
            force: Whether to force removal.
            files: Whether to remove files as well.

        Returns:
            Always True (raises exception otherwise).

        Raises:
            ClientException: When removal failed.
        """  # noqa: DAR401,DAR402 (raise result)
        result = self.api.remove([self], force=force, files=files)[0]
        if not result:
            raise result  # type: ignore  # we know it's a ClientException
        return True

    def pause(self, force: bool = False) -> bool:
        """
        Pause the download.

        Arguments:
            force: Whether to force pause (don't contact servers).

        Returns:
            Always True (raises exception otherwise).

        Raises:
            ClientException: When pausing failed.
        """  # noqa: DAR401,DAR402 (raise result)
        result = self.api.pause([self], force=force)[0]
        if not result:
            raise result  # type: ignore  # we know it's a ClientException
        return True

    def resume(self) -> bool:
        """
        Resume the download.

        Returns:
            Always True (raises exception otherwise).

        Raises:
            ClientException: When resuming failed.
        """  # noqa: DAR401,DAR402 (raise result)
        result = self.api.resume([self])[0]
        if not result:
            raise result  # type: ignore  # we know it's a ClientException
        return True

    def purge(self) -> bool:
        """
        Purge itself from the results.

        Returns:
            Success or failure of the operation.
        """
        return self.api.client.remove_download_result(self.gid) == "OK"

    def move_files(self, to_directory: PathOrStr, force: bool = False) -> bool:
        """
        Move downloaded files to another directory.

        Arguments:
            to_directory: The target directory to move files to.
            force: Whether to move files even if download is not complete.

        Returns:
            Success or failure of the operation.
        """
        return self.api.move_files([self], to_directory, force)[0]

    def copy_files(self, to_directory: PathOrStr, force: bool = False) -> bool:
        """
        Copy downloaded files to another directory.

        Arguments:
            to_directory: The target directory to copy files into.
            force: Whether to move files even if download is not complete.

        Returns:
            Success or failure of the operation.
        """
        return self.api.copy_files([self], to_directory, force)[0]



"""
This module defines the Stats class.

It holds information retrieved with the `get_global_stat` method of the client.
"""

from aria2p.utils import human_readable_bytes


class Stats:
    """This class holds information retrieved with the `get_global_stat` method of the client."""

    def __init__(self, struct: dict) -> None:
        """
        Initialize the object.

        Arguments:
            struct: A dictionary Python object returned by the JSON-RPC client.
        """
        self._struct = struct or {}

    @property
    def download_speed(self) -> int:
        """
        Overall download speed (byte/sec).

        Returns:
            The overall download speed in bytes per second.
        """
        return int(self._struct["downloadSpeed"])

    def download_speed_string(self, human_readable: bool = True) -> str:
        """
        Return the download speed as string.

        Arguments:
            human_readable: Return in human readable format or not.

        Returns:
            The download speed string.
        """
        if human_readable:
            return human_readable_bytes(self.download_speed, delim=" ", postfix="/s")
        return str(self.download_speed) + " B/s"

    @property
    def upload_speed(self) -> int:
        """
        Overall upload speed (byte/sec).

        Returns:
            The overall upload speed in bytes per second.
        """
        return int(self._struct["uploadSpeed"])

    def upload_speed_string(self, human_readable: bool = True) -> str:
        """
        Return the upload speed as string.

        Arguments:
            human_readable: Return in human readable format or not.

        Returns:
            The upload speed string.
        """
        if human_readable:
            return human_readable_bytes(self.upload_speed, delim=" ", postfix="/s")
        return str(self.upload_speed) + " B/s"

    @property
    def num_active(self) -> int:
        """
        Return the number of active downloads.

        Returns:
            The number of active downloads.
        """
        return int(self._struct["numActive"])

    @property
    def num_waiting(self) -> int:
        """
        Return the number of waiting downloads.

        Returns:
            The number of waiting downloads.
        """
        return int(self._struct["numWaiting"])

    @property
    def num_stopped(self) -> int:
        """
        Return the number of stopped downloads in the current session.

        This value is capped by the [`--max-download-result`][aria2p.options.Options.max_download_result] option.

        Returns:
            The number of stopped downloads in the current session (capped).
        """
        return int(self._struct["numStopped"])

    @property
    def num_stopped_total(self) -> int:
        """
        Return the number of stopped downloads in the current session.

        This value is not capped by the [`--max-download-result`][aria2p.options.Options.max_download_result] option.

        Returns:
            The number of stopped downloads in the current session (not capped).
        """
        return int(self._struct["numStoppedTotal"])



"""Complex type annotations."""

from pathlib import Path
from typing import List, Tuple, Union

import aria2p

PathOrStr = Union[Path, str]
OptionsType = Union["aria2p.options.Options", dict]
OperationResult = Union[bool, "aria2p.client.ClientException"]
CallsType = List[Tuple[str, List[str], Union[str, int]]]
Multicalls2Type = List[Tuple[str, List[str]]]



# Why does this file exist, and why not put this in `__main__`?
#
# You might be tempted to import things from `__main__` later,
# but that will cause problems: the code will get executed twice:
#
# - When you run `python -m aria2p` python will execute
#   `__main__.py` as a script. That means there won't be any
#   `aria2p.__main__` in `sys.modules`.
# - When you import `__main__` it will get executed again (as a module) because
#   there's no `aria2p.__main__` in `sys.modules`.

"""Module that contains the command line application."""

import argparse
import json
import sys
from importlib import util as importlib_util
from pathlib import Path
from typing import List, Optional, Union

import requests
from loguru import logger

from aria2p import enable_logger
from aria2p.api import API
from aria2p.client import DEFAULT_HOST, DEFAULT_PORT, DEFAULT_TIMEOUT, Client, ClientException
from aria2p.types import PathOrStr
from aria2p.utils import read_lines

try:
    from aria2p.interface import Interface
except ImportError:
    Interface = None  # type: ignore  # noqa: WPS440 (variable overlap)


# ============ MAIN FUNCTION ============ #
def main(args: Optional[List[str]] = None) -> int:
    """
    Run the main program.

    This function is executed when you type `aria2p` or `python -m aria2p`.

    Arguments:
        args: Arguments passed from the command line.

    Returns:
        An exit code.
    """
    parser = get_parser()
    opts = parser.parse_args(args=args)
    kwargs = opts.__dict__  # noqa: WPS609 (special attribute)

    log_level = kwargs.pop("log_level")
    log_path = kwargs.pop("log_path")

    if log_path:
        log_path = Path(log_path)
        if log_path.is_dir():
            log_path = log_path / "aria2p-{time}.log"
        enable_logger(sink=log_path, level=log_level or "WARNING")
    elif log_level:
        enable_logger(sink=sys.stderr, level=log_level)

    logger.debug("Checking arguments")
    check_args(parser, opts)

    logger.debug("Instantiating API")
    api = API(
        Client(
            host=kwargs.pop("host"),
            port=kwargs.pop("port"),
            secret=kwargs.pop("secret"),
            timeout=kwargs.pop("client_timeout"),
        ),
    )

    logger.info(f"API instantiated: {api!r}")

    # Warn if no aria2 daemon process seems to be running
    logger.debug("Testing connection")
    try:
        api.client.get_version()
    except requests.ConnectionError as error:
        print(f"[ERROR] {error}", file=sys.stderr)
        print(file=sys.stderr)
        print("Please make sure that an instance of aria2c is running with RPC mode enabled,", file=sys.stderr)
        print("and that you have provided the right host, port and secret token.", file=sys.stderr)
        print("More information at https://aria2p.readthedocs.io/en/latest.", file=sys.stderr)
        return 2

    subcommands = {
        None: subcommand_top,
        "show": subcommand_show,
        "top": subcommand_top,
        "call": subcommand_call,
        "add": subcommand_add,
        "add-magnets": subcommand_add_magnets,
        "add-torrents": subcommand_add_torrents,
        "add-metalinks": subcommand_add_metalinks,
        "pause": subcommand_pause,
        "stop": subcommand_pause,  # alias for pause
        "resume": subcommand_resume,
        "start": subcommand_resume,  # alias for resume
        "remove": subcommand_remove,
        "rm": subcommand_remove,  # alias for remove
        "del": subcommand_remove,  # alias for remove
        "delete": subcommand_remove,  # alias for remove
        "purge": subcommand_purge,
        "autopurge": subcommand_purge,  # alias for purge
        "autoclear": subcommand_purge,  # alias for purge
        "autoremove": subcommand_purge,  # alias for purge
        "listen": subcommand_listen,
    }

    subcommand = kwargs.pop("subcommand")

    if subcommand:
        logger.debug("Running subcommand " + subcommand)
    try:
        return subcommands[subcommand](api, **kwargs)
    except ClientException as error:  # noqa: WPS440 (variable overlap)
        print(str(error), file=sys.stderr)
        return error.code


def check_args(parser: argparse.ArgumentParser, opts: argparse.Namespace) -> None:  # noqa: WPS231 (complex)
    """
    Additional checks for command line arguments.

    Arguments:
        parser: An argument parser.
        opts: Parsed options.
    """
    subparsers = [
        action
        for action in parser._actions  # noqa: WPS437 (protected attribute)
        if isinstance(action, argparse._SubParsersAction)  # noqa: WPS437
    ][0].choices

    gid_commands = (
        "pause",
        "stop",
        "remove",
        "rm",
        "del",
        "delete",
        "resume",
        "start",
        "autopurge",
        "autoclear",
        "autoremove",
    )

    if opts.subcommand in gid_commands:
        if not opts.do_all and not opts.gids:
            subparsers[opts.subcommand].error("the following arguments are required: gids or --all")
        elif opts.do_all and opts.gids:
            subparsers[opts.subcommand].error("argument -a/--all: not allowed with arguments gids")
    elif opts.subcommand and opts.subcommand.startswith("add"):
        if not opts.uris and not opts.from_file:
            subparsers[opts.subcommand].error("the following arguments are required: uris")


def get_parser() -> argparse.ArgumentParser:
    """
    Return a parser for the command-line options and arguments.

    Returns:
        An argument parser.
    """
    usage = "%(prog)s [GLOBAL_OPTS...] COMMAND [COMMAND_OPTS...]"  # noqa: WPS323 (%-formatting)
    description = "Command-line tool and Python library to interact with an `aria2c` daemon process through JSON-RPC."
    parser = argparse.ArgumentParser(add_help=False, usage=usage, description=description, prog="aria2p")

    main_help = "Show this help message and exit. Commands also accept the -h/--help option."
    subcommand_help = "Show this help message and exit."

    global_options = parser.add_argument_group(title="Global options")
    global_options.add_argument("-h", "--help", action="help", help=main_help)

    global_options.add_argument(
        "-p",
        "--port",
        dest="port",
        default=DEFAULT_PORT,
        type=int,
        help="Port to use to connect to the remote server.",
    )
    global_options.add_argument(
        "-H",
        "--host",
        dest="host",
        default=DEFAULT_HOST,
        help="Host address for the remote server.",
    )
    global_options.add_argument(
        "-s",
        "--secret",
        dest="secret",
        default="",
        help="Secret token to use to connect to the remote server.",
    )
    global_options.add_argument(
        "-L",
        "--log-level",
        dest="log_level",
        default=None,
        help="Log level to use",
        choices=("TRACE", "DEBUG", "INFO", "SUCCESS", "WARNING", "ERROR", "CRITICAL"),
        type=str.upper,
    )
    global_options.add_argument(
        "-P",
        "--log-path",
        dest="log_path",
        default=None,
        help="Log path to use. Can be a directory or a file.",
    )
    global_options.add_argument(
        "-T",
        "--client-timeout",
        dest="client_timeout",
        default=DEFAULT_TIMEOUT,
        type=float,
        help=f"Timeout in seconds for requests to the remote server. Floats supported. Default: {DEFAULT_TIMEOUT}.",
    )

    # ========= SUBPARSERS ========= #
    subparsers = parser.add_subparsers(dest="subcommand", title="Commands", metavar="", prog="aria2p")

    def subparser(command: str, text: str, **kwargs) -> argparse.ArgumentParser:  # noqa: WPS430 (nested function)
        sub = subparsers.add_parser(command, add_help=False, help=text, description=text, **kwargs)
        sub.add_argument("-h", "--help", action="help", help=subcommand_help)
        return sub

    add_parser = subparser("add", "Add downloads with URIs/Magnets/torrents/Metalinks.")
    add_magnets_parser = subparser("add-magnets", "Add downloads with Magnet URIs.", aliases=["add-magnet"])
    add_metalinks_parser = subparser("add-metalinks", "Add downloads with Metalink files.", aliases=["add-metalink"])
    add_torrents_parser = subparser("add-torrents", "Add downloads with torrent files.", aliases=["add-torrent"])
    subparser(
        "purge",
        "Automatically purge completed/removed/failed downloads.",
        aliases=["autoclear", "autopurge", "autoremove"],
    )
    call_parser = subparser("call", "Call a remote method through the JSON-RPC client.")
    pause_parser = subparser("pause", "Pause downloads.", aliases=["stop"])
    remove_parser = subparser("remove", "Remove downloads.", aliases=["rm", "del", "delete"])
    resume_parser = subparser("resume", "Resume downloads.", aliases=["start"])
    subparser("show", "Show the download progression.")
    subparser("top", "Launch the top-like interactive interface.")
    listen_parser = subparser("listen", "Listen to notifications.")

    # ========= CALL PARSER ========= #
    call_parser.add_argument(
        "method",
        help=(
            "The method to call (case insensitive). "
            "Dashes and underscores will be removed so you can use as many as you want, or none. "
            "Prefixes like 'aria2.' or 'system.' are also optional."
        ),
    )
    call_parser_mxg = call_parser.add_mutually_exclusive_group()
    call_parser_mxg.add_argument(
        "-P",
        "--params-list",
        dest="params",
        nargs="+",
        help="Parameters as a list of strings.",
    )
    call_parser_mxg.add_argument(
        "-J",
        "--json-params",
        dest="params",
        help="Parameters as a JSON string. You should always wrap it at least once in an array '[]'.",
    )

    # ========= ADD PARSER ========= #
    add_parser.add_argument("uris", nargs="*", help="The URIs/file-paths to add.")
    add_parser.add_argument("-f", "--from-file", dest="from_file", help="Load URIs from a file.")

    # ========= ADD MAGNET PARSER ========= #
    add_magnets_parser.add_argument("uris", nargs="*", help="The magnet URIs to add.")
    add_magnets_parser.add_argument("-f", "--from-file", dest="from_file", help="Load URIs from a file.")

    # ========= ADD TORRENT PARSER ========= #
    add_torrents_parser.add_argument("torrent_files", nargs="*", help="The paths to the torrent files.")
    add_torrents_parser.add_argument("-f", "--from-file", dest="from_file", help="Load file paths from a file.")

    # ========= ADD METALINK PARSER ========= #
    add_metalinks_parser.add_argument("metalink_files", nargs="*", help="The paths to the metalink files.")
    add_metalinks_parser.add_argument("-f", "--from-file", dest="from_file", help="Load file paths from a file.")

    # ========= PAUSE PARSER ========= #
    pause_parser.add_argument("gids", nargs="*", help="The GIDs of the downloads to pause.")
    pause_parser.add_argument("-a", "--all", action="store_true", dest="do_all", help="Pause all the downloads.")
    pause_parser.add_argument(
        "-f",
        "--force",
        dest="force",
        action="store_true",
        help="Pause without contacting servers first.",
    )

    # ========= RESUME PARSER ========= #
    resume_parser.add_argument("gids", nargs="*", help="The GIDs of the downloads to resume.")
    resume_parser.add_argument("-a", "--all", action="store_true", dest="do_all", help="Resume all the downloads.")

    # ========= REMOVE PARSER ========= #
    remove_parser.add_argument("gids", nargs="*", help="The GIDs of the downloads to remove.")
    remove_parser.add_argument("-a", "--all", action="store_true", dest="do_all", help="Remove all the downloads.")
    remove_parser.add_argument(
        "-f",
        "--force",
        dest="force",
        action="store_true",
        help="Remove without contacting servers first.",
    )

    # ========= LISTEN PARSER ========= #
    listen_parser.add_argument(
        "-c",
        "--callbacks-module",
        dest="callbacks_module",
        help="Path to the Python module defining your notifications callbacks.",
    )
    listen_parser.add_argument(
        "event_types",
        nargs="*",
        help="The types of notifications to process: "
        "start, pause, stop, error, complete or btcomplete. "
        "Example: aria2p listen error btcomplete. "
        "Useful if you want to spawn multiple specialized aria2p listener, "
        "for example one for each type of notification, "
        "but still want to use only one callback file.",
    )
    listen_parser.add_argument(
        "-t",
        "--timeout",
        dest="timeout",
        type=float,
        default=5,
        help="Timeout in seconds to use when waiting for data over the WebSocket at each iteration. "
        "Use small values for faster reactivity when stopping to listen.",
    )

    return parser


# ============ SUBCOMMANDS ============ #
def subcommand_show(api: API) -> int:
    """
    Show subcommand.

    Arguments:
        api: The API instance to use.

    Returns:
        int: Always 0.
    """
    downloads = api.get_downloads()

    def print_line(*args):  # noqa: WPS430 (nested function)
        print("{:<17} {:<9} {:>8} {:>12} {:>12} {:>8}  {}".format(*args))  # noqa: P101 (unindexed params)

    print_line("GID", "STATUS", "PROGRESS", "DOWN_SPEED", "UP_SPEED", "ETA", "NAME")

    for download in downloads:
        print_line(
            download.gid,
            download.status,
            download.progress_string(),
            download.download_speed_string(),
            download.upload_speed_string(),
            download.eta_string(),
            download.name,
        )

    return 0


def subcommand_top(api: API) -> int:
    """
    Top subcommand.

    Arguments:
        api: The API instance to use.

    Returns:
        int: Always 0.
    """
    if Interface is None:
        print(
            "The top-interface dependencies are not installed. Try running `pip install aria2p[tui]` to install them.",
            file=sys.stderr,
        )
        return 1

    interface = Interface(api)
    success = interface.run()
    return 0 if success else 1


def subcommand_call(api: API, method: str, params: Union[str, List[str]]) -> int:
    """
    Call subcommand.

    Arguments:
        api: The API instance to use.
        method: Name of the method to call.
        params: Parameters to use when calling method.

    Returns:
        int: Always 0.
    """
    real_method = get_method(method)

    if real_method is None:
        print(f"aria2p: call: Unknown method {method}.", file=sys.stderr)
        print("  Run 'aria2p call listmethods' to list the available methods.", file=sys.stderr)
        return 1

    if isinstance(params, str):
        params = json.loads(params)
    elif params is None:
        params = []

    response = api.client.call(real_method, params)
    print(json.dumps(response))

    return 0


def get_method(name: str) -> Optional[str]:
    """
    Return the actual aria2 method name from a differently formatted name.

    Arguments:
        name: A method name.

    Returns:
        The real method name.
    """
    methods = {}

    for method in Client.METHODS:
        methods[method.lower()] = method
        methods[method.split(".")[1].lower()] = method

    name = name.lower()
    name = name.replace("-", "")
    name = name.replace("_", "")

    return methods.get(name)


def subcommand_add(api: API, uris: List[str] = None, from_file: str = None) -> int:
    """
    Add magnet subcommand.

    Arguments:
        api: The API instance to use.
        uris: The URIs or file-paths to add.
        from_file: Path to the file to read uris from.
            Deprecated: Every URI that is a valid file-path
            and is not a torrent or a metalink is now read as an input file.

    Returns:
        int: 0 if OK else 1.
    """
    uris = uris or []

    if from_file:
        logger.warning(
            "Deprecation warning: every URI that is a valid file-path "
            "and is not a torrent or a metalink is now read as an input file.",
        )

    new_downloads = []

    for uri in uris:
        new_downloads.extend(api.add(uri))

    if new_downloads:
        for new_download in new_downloads:
            print(f"Created download {new_download.gid}")
        return 0

    print("No new download was created", file=sys.stderr)
    return 1


def subcommand_add_magnets(api: API, uris: List[str] = None, from_file: str = None) -> int:
    """
    Add magnet subcommand.

    Arguments:
        api: The API instance to use.
        uris: The URIs of the magnets.
        from_file: Path to the file to read uris from.

    Returns:
        int: Always 0.
    """
    ok = True

    if not uris:
        uris = []

    if from_file:
        try:
            uris.extend(read_lines(from_file))
        except OSError:
            print(f"Cannot open file: {from_file}", file=sys.stderr)
            ok = False

    for uri in uris:
        new_download = api.add_magnet(uri)
        print(f"Created download {new_download.gid}")

    return 0 if ok else 1


def subcommand_add_torrents(api: API, torrent_files: List[str] = None, from_file: str = None) -> int:
    """
    Add torrent subcommand.

    Arguments:
        api: The API instance to use.
        torrent_files: The paths to the torrent files.
        from_file: Path to the file to read torrent files paths from.

    Returns:
        int: Always 0.
    """
    ok = True

    if not torrent_files:
        torrent_files = []

    if from_file:
        try:
            torrent_files.extend(read_lines(from_file))
        except OSError:
            print(f"Cannot open file: {from_file}", file=sys.stderr)
            ok = False

    for torrent_file in torrent_files:
        new_download = api.add_torrent(torrent_file)
        print(f"Created download {new_download.gid}")

    return 0 if ok else 1


def subcommand_add_metalinks(api: API, metalink_files: List[str] = None, from_file: str = None) -> int:
    """
    Add metalink subcommand.

    Arguments:
        api: The API instance to use.
        metalink_files: The paths to the metalink files.
        from_file: Path to the file to metalink files paths from.

    Returns:
        int: 0 if OK else 1.
    """
    ok = True

    if not metalink_files:
        metalink_files = []

    if from_file:
        try:
            metalink_files.extend(read_lines(from_file))
        except OSError:
            print(f"Cannot open file: {from_file}", file=sys.stderr)
            ok = False

    for metalink_file in metalink_files:
        new_downloads = api.add_metalink(metalink_file)
        for download in new_downloads:
            print(f"Created download {download.gid}")

    return 0 if ok else 1


def subcommand_pause(api: API, gids: List[str] = None, do_all: bool = False, force: bool = False) -> int:
    """
    Pause subcommand.

    Arguments:
        api: The API instance to use.
        gids: The GIDs of the downloads to pause.
        do_all: Pause all downloads if True.
        force: Force pause or not (see API.pause).

    Returns:
        int: 0 if all success, 1 if one failure.
    """
    if do_all:
        if api.pause_all(force=force):
            return 0
        return 1

    try:
        downloads = api.get_downloads(gids)
    except ClientException as error:
        print(str(error), file=sys.stderr)
        return 1

    result = api.pause(downloads, force=force)

    if all(result):
        return 0

    for item in result:
        if isinstance(item, ClientException):
            print(item, file=sys.stderr)

    return 1


def subcommand_resume(api: API, gids: List[str] = None, do_all: bool = False) -> int:
    """
    Resume subcommand.

    Arguments:
        api: The API instance to use.
        gids: The GIDs of the downloads to resume.
        do_all: Pause all downloads if True.

    Returns:
        int: 0 if all success, 1 if one failure.
    """
    if do_all:
        if api.resume_all():
            return 0
        return 1

    try:
        downloads = api.get_downloads(gids)
    except ClientException as error:
        print(str(error), file=sys.stderr)
        return 1

    result = api.resume(downloads)

    if all(result):
        return 0

    for item in result:
        if isinstance(item, ClientException):
            print(item, file=sys.stderr)

    return 1


def subcommand_remove(api: API, gids: List[str] = None, do_all: bool = False, force: bool = False) -> int:
    """
    Remove subcommand.

    Arguments:
        api: The API instance to use.
        gids: The GIDs of the downloads to remove.
        do_all: Pause all downloads if True.
        force: Force pause or not (see API.remove).

    Returns:
        int: 0 if all success, 1 if one failure.
    """
    if do_all:
        if api.remove_all():
            return 0
        return 1

    try:
        downloads = api.get_downloads(gids)
    except ClientException as error:
        print(str(error), file=sys.stderr)
        return 1

    ok = True
    result = api.remove(downloads, force=force)

    if all(result):
        return 0 if ok else 1

    for item in result:
        if isinstance(item, ClientException):
            print(item, file=sys.stderr)

    return 1


def subcommand_purge(api: API) -> int:
    """
    Purge subcommand.

    Arguments:
        api: The API instance to use.

    Returns:
        int: 0 if all success, 1 if one failure.
    """
    if api.autopurge():
        return 0
    return 1


def subcommand_listen(
    api: API,
    callbacks_module: PathOrStr = None,
    event_types: List[str] = None,
    timeout: int = 5,
) -> int:
    """
    Listen subcommand.

    Arguments:
        api: The API instance to use.
        callbacks_module: The path to the module to import, containing the callbacks as functions.
        event_types: The event types to process.
        timeout: The timeout to pass to the WebSocket connection, in seconds.

    Returns:
        int: Always 0.
    """
    if not callbacks_module:
        print("aria2p: listen: Please provide the callback module file path with -c option", file=sys.stderr)
        return 1

    if isinstance(callbacks_module, Path):
        callbacks_module = str(callbacks_module)

    if not event_types:
        event_types = ["start", "pause", "stop", "error", "complete", "btcomplete"]

    spec = importlib_util.spec_from_file_location("aria2p_callbacks", callbacks_module)
    callbacks = importlib_util.module_from_spec(spec)

    if callbacks is None:
        print(f"aria2p: Could not import module file {callbacks_module}", file=sys.stderr)
        return 1

    spec.loader.exec_module(callbacks)

    callbacks_kwargs = {}
    for callback_name in (  # noqa: WPS352 (multiline loop)
        "on_download_start",
        "on_download_pause",
        "on_download_stop",
        "on_download_error",
        "on_download_complete",
        "on_bt_download_complete",
    ):
        if callback_name[3:].replace("download", "").replace("_", "") in event_types:
            callback = getattr(callbacks, callback_name, None)
            if callback:
                callbacks_kwargs[callback_name] = callback

    api.listen_to_notifications(timeout=timeout, handle_signals=True, threaded=False, **callbacks_kwargs)
    return 0



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['aria2p']

package_data = \
{'': ['*']}

install_requires = \
['loguru', 'requests>2.19.1', 'toml>=0.10.2,<0.11.0', 'websocket_client']

extras_require = \
{'tui': ['asciimatics>=1.11.0,<2.0.0']}

entry_points = \
{'console_scripts': ['aria2p = aria2p.cli:main']}

setup_kwargs = {
    'name': 'aria2p',
    'version': '0.10.1',
    'description': 'Command-line tool and library to interact with an aria2c daemon process with JSON-RPC.',
    'long_description': '<!--\nIMPORTANT: This file is generated from the template at \'scripts/templates/README.md\'.\n           Please update the template instead of this file.\n-->\n\n# aria2p\n[![ci](https://github.com/pawamoy/aria2p/workflows/ci/badge.svg)](https://github.com/pawamoy/aria2p/actions?query=workflow%3Aci)\n[![documentation](https://img.shields.io/badge/docs-mkdocs%20material-blue.svg?style=flat)](https://pawamoy.github.io/aria2p/)\n[![pypi version](https://img.shields.io/pypi/v/aria2p.svg)](https://pypi.org/project/aria2p/)\n[![gitter](https://badges.gitter.im/join%20chat.svg)](https://gitter.im/aria2p/community)\n\nCommand-line tool and Python library to interact with an [`aria2c`][1] daemon process through JSON-RPC.\n\n![demo](https://user-images.githubusercontent.com/3999221/72664104-41658180-39fa-11ea-838e-022ed29d8c0b.gif)\n\nTo avoid confusion:\n\n- [*aria2*][1] is a lightweight multi-protocol & multi-source, cross platform download utility operated in command-line.\nIt supports HTTP/HTTPS, FTP, SFTP, BitTorrent and Metalink.\n- `aria2c` is the name of the command-line executable provided by *aria2*. It can act as a daemon.\n- `aria2p` (`p` for Python) is a command-line client that can interact with an `aria2c` daemon.\n  It is not an official client. There are other Python packages allowing you to interact with an `aria2c` daemon.\n  These other packages do not offer enough usability (in my opinion), this is why I\'m developing `aria2p`.\n\n**Purpose**: `aria2c` can run in the foreground, for one-time downloads, or in the background, as a daemon.\nThis is where `aria2p` intervenes: when an instance of `aria2c` is running in the background,\n`aria2p` will be able to communicate with it to add downloads to the queue, remove, pause or resume them, etc.\n\nIn order for `aria2p` to be able to communicate with the `aria2c` process, RPC mode must be enabled\nwith the `--enable-rpc` option of `aria2c`. RPC stands for [Remote Procedure Call][2].\nAlthough `aria2c` supports both JSON-RPC and XML-RPC protocols, `aria2p` **works with JSON only** (not XML).\nMore information about how to configure `aria2c` to run as a daemon with RPC mode enabled\ncan be found in the [Configuration section][conf doc] of the documentation.\n\n[conf doc]: https://aria2p.readthedocs.io/en/latest/configuration.html\n\n**Table of contents**\n\n- [Requirements](#requirements)\n- [Installation](#installation)\n- [Usage as a library](#usage-as-a-library)\n- [Usage on the command line](#usage-command-line)\n- [Troubleshoot](#troubleshoot)\n- [Support](#support)\n\n\n## Requirements\n\n`aria2p` requires Python 3.6 or above.\n\n<details>\n<summary>To install Python 3.6, I recommend using <a href="https://github.com/pyenv/pyenv"><code>pyenv</code></a>.</summary>\n\n```bash\n# install pyenv\ngit clone https://github.com/pyenv/pyenv ~/.pyenv\n\n# setup pyenv (you should also put these three lines in .bashrc or similar)\nexport PATH="${HOME}/.pyenv/bin:${PATH}"\nexport PYENV_ROOT="${HOME}/.pyenv"\neval "$(pyenv init -)"\n\n# install Python 3.6\npyenv install 3.6.12\n\n# make it available globally\npyenv global system 3.6.12\n```\n</details>\n\nYou must also install [*aria2*][1]. On systems with `apt-get`:\n\n```bash\nsudo apt-get install aria2\n```\n\n[1]: https://github.com/aria2/aria2\n[2]: https://en.wikipedia.org/wiki/Remote_procedure_call\n\n## Installation\n\nWith `pip`:\n\n```bash\npython3.6 -m pip install aria2p[tui]\n```\n\nWith [`pipx`](https://github.com/pipxproject/pipx):\n\n```bash\npython3.6 -m pip install --user pipx\n\npipx install --python python3.6 aria2p[tui]\n```\n\nThe `tui` extra is needed for the interactive interface. If you don\'t need the interface (for example when you are\nwriting a Python package with a dependency to aria2p), simply install `aria2p` without any extra.\n\n## Usage (as a library)\n\n**This library is still a work in progress. More examples will be added later.\nIn the meantime, you can read the [Reference section](https://aria2p.readthedocs.io/en/latest/reference.html) on the official documentation.**\n\n```python\nimport aria2p\n\n# initialization, these are the default values\naria2 = aria2p.API(\n    aria2p.Client(\n        host="http://localhost",\n        port=6800,\n        secret=""\n    )\n)\n\n# list downloads\ndownloads = aria2.get_downloads()\n\nfor download in downloads:\n    print(download.name, download.download_speed)\n\n# add downloads\nmagnet_uri = "magnet:?xt=urn:..."\n\ndownload = aria2.add_magnet(magnet_uri)\n```\n\n## Usage (command-line)\n\n```\nusage: aria2p [GLOBAL_OPTS...] COMMAND [COMMAND_OPTS...]\n\nCommand-line tool and Python library to interact with an `aria2c` daemon\nprocess through JSON-RPC.\n\nGlobal options:\n  -h, --help            Show this help message and exit. Commands also accept\n                        the -h/--help option.\n  -p PORT, --port PORT  Port to use to connect to the remote server.\n  -H HOST, --host HOST  Host address for the remote server.\n  -s SECRET, --secret SECRET\n                        Secret token to use to connect to the remote server.\n  -L {TRACE,DEBUG,INFO,SUCCESS,WARNING,ERROR,CRITICAL}, --log-level {TRACE,DEBUG,INFO,SUCCESS,WARNING,ERROR,CRITICAL}\n                        Log level to use\n  -P LOG_PATH, --log-path LOG_PATH\n                        Log path to use. Can be a directory or a file.\n  -T CLIENT_TIMEOUT, --client-timeout CLIENT_TIMEOUT\n                        Timeout in seconds for requests to the remote server.\n                        Floats supported. Default: 60.0.\n\nCommands:\n  \n    add                 Add downloads with URIs/Magnets/torrents/Metalinks.\n    add-magnets (add-magnet)\n                        Add downloads with Magnet URIs.\n    add-metalinks (add-metalink)\n                        Add downloads with Metalink files.\n    add-torrents (add-torrent)\n                        Add downloads with torrent files.\n    autopurge (autoclear)\n                        Automatically purge completed/removed/failed\n                        downloads.\n    call                Call a remote method through the JSON-RPC client.\n    pause (stop)        Pause downloads.\n    remove (rm, del, delete)\n                        Remove downloads.\n    resume (start)      Resume downloads.\n    show                Show the download progression.\n    top                 Launch the top-like interactive interface.\n    listen              Listen to notifications.\n\n```\n\nCalling `aria2p` without any arguments will by default call the `top` command,\nwhich is a console interactive interface.\n\nCommands:\n\n\n- [`add`](#add)\n- [`add-magnets`](#add-magnets)\n- [`add-metalinks`](#add-metalinks)\n- [`add-torrents`](#add-torrents)\n- [`autopurge`](#autopurge)\n- [`call`](#call)\n- [`listen`](#listen)\n- [`pause`](#pause)\n- [`remove`](#remove)\n- [`resume`](#resume)\n- [`show`](#show)\n- [`top`](#top)\n\n\n---\n\n### `add`\n\n```\nusage: aria2p add [-h] [-f FROM_FILE] [uris [uris ...]]\n\nAdd downloads with URIs/Magnets/torrents/Metalinks.\n\npositional arguments:\n  uris                  The URIs/file-paths to add.\n\noptional arguments:\n  -h, --help            Show this help message and exit.\n  -f FROM_FILE, --from-file FROM_FILE\n                        Load URIs from a file.\n\n```\n\n\n\n---\n\n### `add-magnets`\n\n```\nusage: aria2p add-magnets [-h] [-f FROM_FILE] [uris [uris ...]]\n\nAdd downloads with Magnet URIs.\n\npositional arguments:\n  uris                  The magnet URIs to add.\n\noptional arguments:\n  -h, --help            Show this help message and exit.\n  -f FROM_FILE, --from-file FROM_FILE\n                        Load URIs from a file.\n\n```\n\n\n\n---\n\n### `add-metalinks`\n\n```\nusage: aria2p add-metalinks [-h] [-f FROM_FILE]\n                            [metalink_files [metalink_files ...]]\n\nAdd downloads with Metalink files.\n\npositional arguments:\n  metalink_files        The paths to the metalink files.\n\noptional arguments:\n  -h, --help            Show this help message and exit.\n  -f FROM_FILE, --from-file FROM_FILE\n                        Load file paths from a file.\n\n```\n\n\n\n---\n\n### `add-torrents`\n\n```\nusage: aria2p add-torrents [-h] [-f FROM_FILE]\n                           [torrent_files [torrent_files ...]]\n\nAdd downloads with torrent files.\n\npositional arguments:\n  torrent_files         The paths to the torrent files.\n\noptional arguments:\n  -h, --help            Show this help message and exit.\n  -f FROM_FILE, --from-file FROM_FILE\n                        Load file paths from a file.\n\n```\n\n\n\n---\n\n### `autopurge`\n\n```\nusage: aria2p autopurge [-h]\n\nAutomatically purge completed/removed/failed downloads.\n\noptional arguments:\n  -h, --help  Show this help message and exit.\n\n```\n\n\n\n---\n\n### `call`\n\n```\nusage: aria2p call [-h] [-P PARAMS [PARAMS ...] | -J PARAMS] method\n\nCall a remote method through the JSON-RPC client.\n\npositional arguments:\n  method                The method to call (case insensitive). Dashes and\n                        underscores will be removed so you can use as many as\n                        you want, or none. Prefixes like \'aria2.\' or \'system.\'\n                        are also optional.\n\noptional arguments:\n  -h, --help            Show this help message and exit.\n  -P PARAMS [PARAMS ...], --params-list PARAMS [PARAMS ...]\n                        Parameters as a list of strings.\n  -J PARAMS, --json-params PARAMS\n                        Parameters as a JSON string. You should always wrap it\n                        at least once in an array \'[]\'.\n\n```\n\nAs explained in the help text,\nthe `method` can be the exact method name,\nor just the name without the prefix.\nIt is case-insensitive, and dashes and underscores will be removed.\n\nThe following are all equivalent:\n- `aria2.addUri`\n- `aria2.adduri`\n- `addUri`\n- `ADDURI`\n- `aria2.ADD-URI`\n- `add_uri`\n- `A-d_D-u_R-i` (yes it\'s valid)\n- `A---R---I---A---2.a__d__d__u__r__i` (I think you got it)\n- and even more ugly forms...\n\n#### Examples\nList all available methods.\n*This example uses [`jq`](https://github.com/stedolan/jq).*\n```console\n$ aria2p call listmethods | jq\n[\n  "aria2.addUri",\n  "aria2.addTorrent",\n  "aria2.getPeers",\n  "aria2.addMetalink",\n  "aria2.remove",\n  "aria2.pause",\n  "aria2.forcePause",\n  "aria2.pauseAll",\n  "aria2.forcePauseAll",\n  "aria2.unpause",\n  "aria2.unpauseAll",\n  "aria2.forceRemove",\n  "aria2.changePosition",\n  "aria2.tellStatus",\n  "aria2.getUris",\n  "aria2.getFiles",\n  "aria2.getServers",\n  "aria2.tellActive",\n  "aria2.tellWaiting",\n  "aria2.tellStopped",\n  "aria2.getOption",\n  "aria2.changeUri",\n  "aria2.changeOption",\n  "aria2.getGlobalOption",\n  "aria2.changeGlobalOption",\n  "aria2.purgeDownloadResult",\n  "aria2.removeDownloadResult",\n  "aria2.getVersion",\n  "aria2.getSessionInfo",\n  "aria2.shutdown",\n  "aria2.forceShutdown",\n  "aria2.getGlobalStat",\n  "aria2.saveSession",\n  "system.multicall",\n  "system.listMethods",\n  "system.listNotifications"\n]\n```\n\nList the GIDs (identifiers) of all active downloads.\n*Note that we must give the parameters as a JSON string.*\n```console\n$ aria2p call tellactive -J \'[["gid"]]\'\n[{"gid": "b686cad55029d4df"}, {"gid": "4b39a1ad8fd94e26"}, {"gid": "9d331cc4b287e5df"}, {"gid": "8c9de0df753a5195"}]\n```\n\nPause a download using its GID.\n*Note that when a single string argument is required, it can be passed directly with `-P`.*\n```console\n$ aria2p call pause -P b686cad55029d4df\n"b686cad55029d4df"\n```\n\nAdd a download using magnet URIs.\n*This example uses `jq -r` to remove the quotation marks around the result.*\n```console\n$ aria2p call adduri -J \'[["magnet:?xt=urn:..."]]\' | jq -r\n4b39a1ad8fd94e26f\n```\n\nPurge download results (remove completed downloads from the list).\n```console\n$ aria2p call purge_download_result\n"OK"\n```\n\n---\n\n### `listen`\n\n```\nusage: aria2p listen [-h] [-c CALLBACKS_MODULE] [-t TIMEOUT]\n                     [event_types [event_types ...]]\n\nListen to notifications.\n\npositional arguments:\n  event_types           The types of notifications to process: start, pause,\n                        stop, error, complete or btcomplete. Example: aria2p\n                        listen error btcomplete. Useful if you want to spawn\n                        multiple specialized aria2p listener, for example one\n                        for each type of notification, but still want to use\n                        only one callback file.\n\noptional arguments:\n  -h, --help            Show this help message and exit.\n  -c CALLBACKS_MODULE, --callbacks-module CALLBACKS_MODULE\n                        Path to the Python module defining your notifications\n                        callbacks.\n  -t TIMEOUT, --timeout TIMEOUT\n                        Timeout in seconds to use when waiting for data over\n                        the WebSocket at each iteration. Use small values for\n                        faster reactivity when stopping to listen.\n\n```\n\n\n\n---\n\n### `pause`\n\n```\nusage: aria2p pause [-h] [-a] [-f] [gids [gids ...]]\n\nPause downloads.\n\npositional arguments:\n  gids         The GIDs of the downloads to pause.\n\noptional arguments:\n  -h, --help   Show this help message and exit.\n  -a, --all    Pause all the downloads.\n  -f, --force  Pause without contacting servers first.\n\n```\n\n\n\n---\n\n### `remove`\n\n```\nusage: aria2p remove [-h] [-a] [-f] [gids [gids ...]]\n\nRemove downloads.\n\npositional arguments:\n  gids         The GIDs of the downloads to remove.\n\noptional arguments:\n  -h, --help   Show this help message and exit.\n  -a, --all    Remove all the downloads.\n  -f, --force  Remove without contacting servers first.\n\n```\n\n\n\n---\n\n### `resume`\n\n```\nusage: aria2p resume [-h] [-a] [gids [gids ...]]\n\nResume downloads.\n\npositional arguments:\n  gids        The GIDs of the downloads to resume.\n\noptional arguments:\n  -h, --help  Show this help message and exit.\n  -a, --all   Resume all the downloads.\n\n```\n\n\n\n---\n\n### `show`\n\n```\nusage: aria2p show [-h]\n\nShow the download progression.\n\noptional arguments:\n  -h, --help  Show this help message and exit.\n\n```\n\n\n\n---\n\n### `top`\n\n```\nusage: aria2p top [-h]\n\nLaunch the top-like interactive interface.\n\noptional arguments:\n  -h, --help  Show this help message and exit.\n\n```\n\n\n\n\n## Troubleshooting\n\n- Error outputs like below when using `aria2p` as a Python library:\n\n  ```\n  requests.exceptions.ConnectionError: HTTPConnectionPool(host=\'localhost\', port=6800): Max retries exceeded with url: /jsonrpc (Caused by NewConnectionError(\'<urllib3.connection.HTTPConnection object at 0x1115b1908>: Failed to establish a new connection: [Errno 61] Connection refused\',))\n  ```\n\n  Solution: `aria2c` needs to be up and running first.\n\n## Support\n\nTo support me as an open-source software author,\nconsider donating or be a supporter through one of the following platforms\n\n- [GitHub](https://github.com/sponsors/pawamoy)\n- [Ko-fi](https://www.ko-fi.com/pawamoy)\n- [Liberapay](https://liberapay.com/pawamoy/)\n- [Patreon](https://www.patreon.com/pawamoy)\n- [Paypal](https://www.paypal.me/pawamoy)\n\nThank you!',
    'author': 'Timothée Mazzucotelli',
    'author_email': 'pawamoy@pm.me',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/pawamoy/aria2p',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



"""
aria2p package.

Command-line tool and library to interact with an aria2c daemon process with JSON-RPC.
"""

import sys

from loguru import logger

from aria2p.api import API
from aria2p.client import Client, ClientException
from aria2p.downloads import BitTorrent, Download, File
from aria2p.options import Options
from aria2p.stats import Stats

logger.disable("aria2p")


def enable_logger(sink=sys.stderr, level="WARNING"):
    """
    Enable the logging of messages.

    Configure the `logger` variable imported from `loguru`.

    Arguments:
        sink (file): An opened file pointer, or stream handler. Default to standard error.
        level (str): The log level to use. Possible values are TRACE, DEBUG, INFO, WARNING, ERROR, CRITICAL.
            Default to WARNING.
    """
    logger.remove()
    logger.configure(handlers=[{"sink": sink, "level": level}])
    logger.enable("aria2p")


__all__ = [  # noqa: WPS410 (the only __variable__ we use)
    "API",
    "ClientException",
    "Client",
    "Download",
    "BitTorrent",
    "File",
    "Options",
    "Stats",
    "enable_logger",
]
